import sys
import inspect
from cloudmesh.common.util import readfile


class CloudmeshPlugin:

    @staticmethod
    def inheritors(klass):
        """
        returns the inheritors of a class if it is loaded

        :return: a set of classes
        """
        subclasses = set()
        work = [klass]
        while work:
            parent = work.pop()
            for child in parent.__subclasses__():
                if child not in subclasses:
                    subclasses.add(child)
                    work.append(child)
        return subclasses

    @staticmethod
    def find_PluginCommands_in_sys():

        modules = sys.modules.keys()
        plugins = []
        for module in modules:
            if module.startswith("cloudmesh."):
                try:

                    _class = __import__(module)

                    print("check", module, _class)

                    filename = inspect.getfile(_class)
                    content = readfile(filename)
                    if "PluginCommand" in content and "cloudmesh.shell.command" in content:
                        entry = {
                            "module": module,
                            "filename": filename
                        }
                        plugins.append(entry)
                except Exception as e:
                    print(e)
                    pass

        return plugins



import yaml
import requests
import os


class PluginManager(object):

    def __init(self):
        self.data = None

    def load(self):
        """
        TODO: DEPRECATED:

        loads the plugins form the plugin.yml file in github

        :return:
        """
        url = 'https://raw.githubusercontent.com/cloudmesh/cloudmesh-cmd5/install/plugins.yml'
        r = requests.get(url)
        self.data = yaml.load(r.text)
        for key in self.data['plugins']:
            entry = self.data['plugins'][key]
            entry['description'] = entry['description'].strip()

    def uninstall(self, name):
        """
        pip uninstalls the package with the specified name
        :param name: The package name
        :return:
        """
        plugin = "cloudmesh." + name
        print("Uninstalling:", )
        os.system("pip uninstall {name}".format(name=plugin))

    def pip_install(self, name):
        """
        pip installs the package with the specified name
        :param name: The package name
        :return:
        """
        plugin = "cloudmesh." + name
        print("Installing:", )
        os.system("pip install {name}".format(name=plugin))

    def source_install(self, name):
        """
        TODO: DEPRECATED

        :param name:
        :return:
        """
        pass



import importlib
import inspect
import os
import pkgutil
import pydoc
import subprocess
import sys
import textwrap
from cmd import Cmd

import cloudmesh
import cloudmesh.common
from cloudmesh.common.Printer import Printer
from cloudmesh.common.Shell import Shell
from cloudmesh.common.StopWatch import StopWatch
from cloudmesh.common.console import Console
from cloudmesh.common.default import Default
from cloudmesh.common.dotdict import dotdict
from cloudmesh.common.error import Error
from cloudmesh.common.util import path_expand
from cloudmesh.common.util import readfile
from cloudmesh.common.variables import Variables
from cloudmesh.shell.command import PluginCommand
from cloudmesh.shell.command import command, basecommand
from cloudmesh.shell.plugin import PluginManager


# import cloudmesh.plugin


def iter_namespace(ns_pkg):
    return pkgutil.iter_modules(ns_pkg.__path__, ns_pkg.__name__ + ".")


# cloudmesh_plugins = {
#    name: importlib.import_module(name)
#    for finder, name, ispkg
#    in iter_namespace(cloudmesh.plugin)
# }


def import_class(cl):
    d = cl.rfind(".")
    classname = cl[d + 1:len(cl)]
    m = __import__(cl[0:d], globals(), locals(), [classname])
    return getattr(m, classname)


def get_class(str):
    return getattr(sys.modules[__name__], str)


class Plugin(object):
    """
    Some simple methods to manage dynamic namespace plugins for cloudmesh.
    """

    @staticmethod
    def list():
        subclasses = set()
        work = [PluginCommand]
        while work:
            parent = work.pop()
            for child in parent.__subclasses__():
                if child not in subclasses and child.__name__ not in ["CMShell",
                                                                      "CommandProxyClass"]:
                    subclasses.add(child)
                    work.append(child)
        return subclasses

    @classmethod
    def modules(cls):
        """
        list of cloudmesh modules in the cloudmesh namespace
        :return: list of modules
        """

        module_list = []
        package = cloudmesh
        for importer, modname, ispkg in pkgutil.walk_packages(
                path=package.__path__,
                prefix=package.__name__ + '.',
                onerror=lambda x: None):
            module_list.append(modname)
        return module_list

    @classmethod
    def classes(cls):
        """
        list of the commands in the cloudmesh namespace under cloudmesh.command

        :return: list of the commands
        """
        module_list = cls.modules()
        # VERBOSE(module_list)
        commands = []
        for module_name in module_list:
            if module_name.startswith('cloudmesh.') and '.command.' in module_name:
                commands.append(module_name)

        return commands

    @classmethod
    def name(cls, command):
        """
        creates a name for a modules starting with do_
        :param command: returns a tuple with the module location and tge do_function
        :return:
        """
        command_name = "do_" + command

        class_name = "cloudmesh." + command + ".command." + command + "." \
                     + command.capitalize() + "Command"

        return class_name, command_name

    @classmethod
    def class_name(cls, command):
        """
        creates the default filename in which the module is defined
        :param command:  the name of the command
        :return: cloudmesh.ext.command.<command>+command.<Command>
        """
        return "cloudmesh." + command + ".command." + command

    @classmethod
    def load(cls, commands=None):
        """
        :param commands: If None the commands will be found from import cloudmesh
                         Otherwise the commands can be explicitly specified with

                          commands = [
                            'cloudmesh.bar.command.bar.BarCommand',
                            'cloudmesh.foo.command.foo.FooCommand',
                            ]
                          A namespace package must exists. Foo and Bar ar just examples

        :return: the classes of the command
        """
        if commands is None:
            classes = cls.classes()
        else:
            classes = commands

        commands = [pydoc.locate(x) for x in classes]

        return commands


selective_loading = True

StopWatch.start("load")
if selective_loading:
    try:
        name = sys.argv[1]

        if name in ["help"]:
            Plugin.load()
        if name in ['banner',
                    'stopwatch',
                    'clear',
                    'sleep',
                    'echo',
                    'term',
                    'pause']:
            name = Plugin.class_name("terminal")
            Plugin.load([name])
        else:
            name = Plugin.class_name(name)
            Plugin.load([name])
    except Exception as e:  # noqa: F841
        Plugin.load()
else:
    Plugin.load()
StopWatch.stop("load")

PluginCommandClasses = type(
    'CommandProxyClass',
    tuple(PluginCommand.__subclasses__()),
    {})

Console.init()


class CMShell(Cmd, PluginCommandClasses):
    """
    The command shell that inherits all commands from PluginCommand
    """
    prompt = 'cms> '
    banner = textwrap.dedent("""
    +-------------------------------------------------------+
    |   ____ _                 _                     _      |
    |  / ___| | ___  _   _  __| |_ __ ___   ___  ___| |__   |
    | | |   | |/ _ \| | | |/ _` | '_ ` _ \ / _ \/ __| '_ \  |
    | | |___| | (_) | |_| | (_| | | | | | |  __/\__ \ | | | |
    |  \____|_|\___/ \__,_|\__,_|_| |_| |_|\___||___/_| |_| |
    +-------------------------------------------------------+
    |                  Cloudmesh CMD5 Shell                 |
    +-------------------------------------------------------+
    """)  # noqa: W605

    def set_debug(self, on):

        def _boolean(value):
            value = str(value).lower()
            if value in ["true", "on"]:
                value = True
            elif value in ["false", "off"]:
                value = False
            else:
                value = False
                Console.error("Value is not boolean")
            return value

        variables = Variables()
        if _boolean(on):
            variables["debug"] = True
            variables["trace"] = True
            variables["verbose"] = '10'
            variables["timer"] = True
        else:
            variables["debug"] = False
            variables["trace"] = False
            variables["verbose"] = '0'
            variables["timer"] = False

        for key in ["debug", "trace", "verbose", "timer"]:
            print(f"{key}={variables[key]}")

    @command
    def do_commands(self, args, arguments):
        """
        ::

          Usage:
                commands

        """
        arguments = dotdict(arguments)

        print("Plugin.list()")
        print(Plugin.list())

        names = self.get_names()

        print()
        print("get_names()")
        print(names)

        print()
        print("dir")
        print(dir(cloudmesh))

    def precmd(self, line):
        print(line)
        StopWatch.start("command")
        return line

    def postcmd(self, stop, line):
        StopWatch.stop("command")

        try:
            variable = Variables()
            if "timer" not in variable:
                variable["timer"] = "False"
            if str(variable["timer"].lower()) in ['on', 'true']:
                command_time = StopWatch.get("command")
                load_time = StopWatch.get("load")
                line_strip = line.strip()
                print(f"Timer: {command_time:.4f}s "
                      f"Load: {load_time:.4f}s "
                      f"{line_strip}")
            variable.close()
        except Exception as e:
            Error.traceback(error=e)

        return stop

    # noinspection PyMethodMayBeStatic
    def replace_vars(self, line):

        # self.update_time()

        variable = Variables()
        newline = line

        if len(variable) is not None:
            for name in variable.data:
                value = str(variable[name])
                newline = newline.replace("$" + name, value)
                newline = newline.replace("var." + name, value)
            for v in os.environ:
                name = v.replace('os.', '')
                if name in newline:
                    value = os.environ[name]
                    newline = newline.replace("os." + v, value)

            default = Default()
            if default is not None:
                for v in default.data:
                    name = "default." + v.replace(",", ".")
                    value = default.data[v]
                    if name in newline:
                        newline = newline.replace(name, value)

                # replace if global is missing

                global_default = default["global"]
                if global_default is not None:
                    for v in global_default:
                        name = "default." + v
                        value = global_default[v]
                        if name in newline:
                            newline = newline.replace(name, value)

            default.close()
            variable.close()

        newline = path_expand(newline)
        return line, newline

    def onecmd(self, line):
        """Interpret the argument as though it had been typed in response
        to the prompt.

        This may be overridden, but should not normally need to be;
        see the precmd() and postcmd() methods for useful execution hooks.
        The return value is a flag indicating whether interpretation of
        commands by the interpreter should stop.

        """
        oldline, line = self.replace_vars(line)

        # -----------------------------
        # print comment lines, but do not execute them
        # -----------------------------
        if line.startswith('#') or line.startswith('//') or line.startswith('/*'):
            print(line)
            return ""

        if line.startswith('!'):
            os.system(line[1:])

            return ""
        # if line is None:
        #    return ""

        # if line.startswith("!"):
        #    line.replace("!", "! ")
        # line = self.var_replacer(line)
        # if line != "hist" and line:
        #    self._hist += [line.strip()]
        # if line.startswith("!") or line.startswith("shell"):
        #    self.do_shell_exec(line[1:])
        #    return ""
        cmd, arg, line = self.parseline(line)

        if line.startswith("$") or line.startswith('var.'):
            line = line.replace("$", "", 1)
            line = line.replace("var.", "", 1)
            print("FIND>", line, "<", sep='')
            variable = Variables()
            print(variable[line])
            variable.close()
            return ""

        # -----------------------------
        # handle empty line
        # -----------------------------
        if not line:
            return self.emptyline()

        # -----------------------------
        # handle file execution
        # -----------------------------
        #
        # this does not yet work
        #
        # if os.path.isfile(line):
        #    print ("... execute", line)
        #    self.do_exec(line)
        #    return ""

        if cmd != '':

            if cmd in ['dryrun']:
                variables = Variables()
                if arg.startswith("="):
                    arg = arg[1:]
                try:
                    value = arg
                    variables.boolean("dryrun", value)
                except IndexError:
                    Console.error("value for dryrun is missing")
                except Exception as e:
                    print(e)

                return ""
            elif cmd in ['debug']:
                if arg.startswith("="):
                    arg = arg[1:]
                try:
                    value = arg
                    self.set_debug(value)
                except IndexError:
                    Console.error("value for dryrun is missing")
                except Exception as e:
                    print(e)

                return ""

            try:
                func = getattr(self, 'do_' + cmd)
                return func(arg)

            except AttributeError as e:

                variables = Variables()
                trace = "T" in variables['trace'] or True
                debug = "T" in variables['debug'] or True

                command_missing = "'CMShell' object has no attribute 'do_{cmd}'".format(
                    cmd=cmd)

                if e.args[0] == command_missing:
                    Console.error(
                        "this command does not exist: '{cmd}'".format(cmd=cmd),
                        traceflag=False)
                else:
                    Error.traceback(error=e, debug=debug, trace=trace)

                # noinspection PyUnusedLocal
                cmd = None
                line = oldline

        return ""

    # noinspection PyUnusedLocal
    @command
    def do_py(self, args, arguments):
        """
        ::

           Usage:
                py PROGRAM...

            Arguments:
                PROGRAM...  the command to be executed

           Description:
                py PROGRAM...  executes the command

        """
        result = eval(args)
        print(result)
        return ""

    # noinspection PyUnusedLocal
    @command
    def do_shell(self, args, arguments):
        """
        ::

           Usage:
                shell COMMAND

            Arguments:
                COMMAND  the command to be executed

           Description:
                shell COMMAND  executes the command

        """
        os.system(str(args))
        return ""

    '''
    #
    # List all commands that start with do
    #
    # noinspection PyMethodOverriding
    @command
    def do_help(self, args, arguments):
        """
        ::
           Usage:
                help

           Description:
                help - List of all registered commands

        """
        print("Help")
        print("====")
        method_list = [n for n, v in inspect.getmembers(self, inspect.ismethod)]
        function_list = [n for n, v in inspect.getmembers(self, inspect.isfunction)]

        commands = method_list + function_list

        for c in sorted(commands):
            if c.startswith("do_"):
                print(c.replace("do_", ""), end=' ')
        print()
        return ""

    '''

    def do_help(self, arg):
        """
        ::

            Usage:
                help
                help COMMAND

            Description:
                List available commands with "help" or detailed help with
                "help COMMAND".

                In addition to the commands it also allows to print the api.
                The classes that help is available for can be listed with

                    cms info
        """
        Plugin.load()

        if arg and "cloudmesh." in arg:
            r = eval(f"help('{arg}')")
            print(r)
            return ""
        elif arg:
            try:
                func = getattr(self, 'help_' + arg)
            except AttributeError:
                try:
                    doc = getattr(self, 'do_' + arg).__doc__
                    if doc:
                        self.stdout.write("Command {arg}\n".format(arg=arg))
                        self.stdout.write(len("Command " + arg) * "=")
                        self.stdout.write("\n")

                        # doc = doc.replace("::\n\n", "")
                        self.stdout.write("%s\n" % str(doc))
                        return
                except AttributeError:
                    pass
                self.stdout.write("%s\n" % str(self.nohelp % (arg,)))
                return
            func()
        else:
            names = self.get_names()
            cmds_doc = []
            cmds_undoc = []
            help_page = {}
            for name in names:
                if name[:5] == 'help_':
                    help_page[name[5:]] = 1
            names.sort()
            # There can be duplicates if routines overridden
            prevname = ''
            for name in names:
                if name[:3] == 'do_':
                    if name == prevname:
                        continue
                    prevname = name
                    cmd = name[3:]
                    if cmd in help_page:
                        cmds_doc.append(cmd)
                        del help_page[cmd]
                    elif getattr(self, name).__doc__:
                        cmds_doc.append(cmd)
                    else:
                        cmds_undoc.append(cmd)

            self.stdout.write("%s\n" % str(self.doc_leader))
            self.print_topics(self.doc_header, cmds_doc, 15, 80)
            self.print_topics(self.misc_header, list(help_page.keys()), 15, 80)
            self.print_topics(self.undoc_header, cmds_undoc, 15, 80)

    def help_help(self):
        """
        ::

            Usage:
                help
                help COMMAND

            Description:
                List available commands with "help" or detailed help with
                "help COMMAND".
        """
        print(textwrap.dedent(self.help_help.__doc__))

    # noinspection PyUnusedLocal
    @command
    def do_info(self, args, arguments):
        """
        ::

          Usage:
            info path
            info commands
            info files [errors] [--output=FORMAT]
            info cloudmesh
            info errors [--trace]
            info sys [KEYWORD] [--plugins] [--output=FORMAT]
            info

          Options:
            --format=FORMAT  the format to print the versions in [default: table]

          Description:

            info
                provides internal info about the shell and its packages

            The format option only works for

                cms info files
        """

        arguments = dotdict(arguments)
        _format = arguments["--output"] or "table"

        module_list = Plugin.modules()

        if arguments.commands:

            commands = Plugin.classes()
            Console.bullets(commands)

        elif arguments.path:

            path_list = cloudmesh.__path__
            Console.bullets(path_list)

        elif arguments.sys:

            keyword = arguments["KEYWORD"]

            all_modules = sys.modules.keys()
            modules = []
            for module in all_modules:
                if keyword:
                    if keyword in module:
                        modules.append(module)
                else:
                    modules.append(module)

            if arguments["--plugins"]:
                plugins = []
                for module in modules:
                    try:
                        _class = PluginManager.get_class(module)
                        filename = inspect.getfile(_class)
                        content = readfile(filename)
                        if "PluginCommand" in content:
                            entry = {
                                "module": module,
                                "filename": filename
                            }
                            plugins.append(entry)
                    except Exception as e:  # noqa: F841
                        # Console.error(str(e))
                        pass
                print(Printer.write(plugins,
                                    order=["module", "filename"],
                                    max_width=256,
                                    output=_format))
            else:
                print("\n".join(modules))

        elif arguments.files or arguments.errors:

            modules = Plugin.modules()
            # longets = 0
            data = []
            for module in modules:
                try:
                    _class = get_class(module)
                    r = inspect.getfile(_class)
                    data.append({
                        "module": module,
                        "directory": r
                    })
                except Exception as e:
                    if arguments.errors:
                        if "got str" not in str(e) and "__version__" not in module:
                            if "module" not in str(e):
                                print()
                                Console.error(f"Problem for {module}")
                                Console.error(str(e),
                                              traceflag=arguments["--trace"])
                                print()
                            else:
                                Console.error(str(e),
                                              traceflag=arguments["--trace"])

            if arguments.files:
                print(Printer.write(data,
                                    order=["module", "directory"],
                                    max_width=256,
                                    output=_format))

        elif arguments.help:

            for name in module_list:
                p = "cloudmesh." + name
                help_string = p + " not found."
                # noinspection PyBroadException
                try:
                    help_string = pydoc.render_doc(p,
                                                   "Help on %s" + "\n" + 79 * "=")
                except Exception as e:  # noqa: F841
                    pass
                print(help_string)

        else:

            Console.bullets(module_list)

    def preloop(self):
        """adds the banner to the preloop"""

        lines = textwrap.dedent(self.banner).splitlines()
        for line in lines:
            # Console.cprint("BLUE", "", line)
            print(line)

    # noinspection PyUnusedLocal,PyPep8Naming,PyMethodMayBeStatic
    def do_EOF(self, args):
        """
        ::

          Usage:
            EOF

          Description:
            Command to the shell to terminate reading a script.
        """
        return True

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def do_quit(self, args):
        """
        ::

          Usage:
            quit

          Description:
            Action to be performed when quit is typed

        """
        return True

    do_q = do_quit

    def emptyline(self):
        return

    # noinspection PyUnusedLocal,PyPep8
    @basecommand
    def do_version(self, args, arguments):
        """
        ::

          Usage:
            version pip [PACKAGE]
            version [--format=FORMAT] [--check=CHECK] [--number]

          Options:
            --format=FORMAT  the format to print the versions in [default: table]
            --check=CHECK    boolean tp conduct an additional check [default: True]

          Description:
            version
                Prints out the version number
            version pip
                Prints the contents of pip list

          Limitations:
            Package names must not have a . in them instead you need to use -
            Thus to query for cloudmesh-cmd5 use

              cms version pip cloudmesh-cmd5

        """

        # print (arguments)
        # print (">", args, "<")

        if arguments["pip"]:
            # noinspection PyBroadException
            try:
                package = arguments["PACKAGE"]

                if package is None:
                    result = Shell.execute('pip', ['list', '--format=columns'],
                                           traceflag=False, witherror=False)
                    print(result)
                else:
                    if "." in package:
                        package = package.replace(".", "-")
                    result = Shell.execute('pip', ['show', package],
                                           traceflag=False, witherror=False)
                    print(result)

            except Exception as e:  # noqa: F841
                result = 'N/A'
            return ""

        python_version, pip_version = Shell.get_python()

        # noinspection PyBroadException
        try:
            git_hash_version = Shell.execute('git', 'log -1 --format=%h',
                                             traceflag=False, witherror=False)
        except:  # noqa: E722
            git_hash_version = 'N/A'

        versions = {
            "python": {
                "name": "python",
                "version": str(python_version)
            },
            "pip": {
                "name": "pip",
                "version": str(pip_version)
            },
            "git": {
                "name": "git hash",
                "version": str(git_hash_version)
            }

        }

        # dynamically check all installed cloudmesh packages and versions
        pipcheck = subprocess.Popen(('pip', 'freeze'), stdout=subprocess.PIPE)
        try:
            # python 3 returns byte sequence so the decode is necessary
            output = subprocess.check_output(('grep', "cloudmesh"),
                                             stdin=pipcheck.stdout).decode(
                "utf-8")
            pkglines = output.strip().splitlines()

            for pkgline in pkglines:
                # print (pkgline)
                if "==" in pkgline:
                    values = pkgline.split("==")
                    pkg = values[0]
                    version = values[1].strip()
                    if pkg != 'cloudmesh-installer':
                        pname = pkg.replace("cloudmesh-", "cloudmesh.")
                        i = importlib.import_module(pname)

                        location = i.__file__

                        try:
                            vlocation = path_expand(
                                os.path.join(os.path.dirname(location),
                                             "__version__.py"))
                            v = readfile(vlocation).split('"')[1].strip()
                        except:  # noqa: E722
                            v = "not found"

                        versions[pkg] = {
                            "name": pkg,
                            "package": pname,
                            "version": version,
                            "source": location,
                            "VERSION": v
                        }
                elif "git" in pkgline and ("installer" not in pkgline) and ("pi_burn" not in pkgline):

                    pkgline = pkgline.replace(
                        "-e git+git@github.com:cloudmesh-community/", "")
                    pkgline = pkgline.replace(
                        "-e git+https://github.com/cloudmesh/", "")
                    pkgline = pkgline.replace("egg=", "")

                    version, pkg = pkgline.split("#")
                    pname = pkg.replace("cloudmesh-", "cloudmesh.")
                    #
                    # The next line needed to be added as for some reason the is an _ here
                    #
                    pname = pkg.replace("cloudmesh_", "cloudmesh.")

                    try:
                        i = importlib.import_module(pname)
                    except:  # noqa: E722
                        continue

                    location = i.__file__

                    try:
                        vlocation = path_expand(
                            os.path.join(os.path.dirname(location),
                                         "__version__.py"))
                        v = readfile(vlocation).split('"')[1].strip()
                    except:  # noqa: E722
                        v = "not found"

                    versions[pkg] = {
                        "name": pkg,
                        "package": pname,
                        "version": version,
                        "source": location,
                        "VERSION": v
                    }
        except subprocess.CalledProcessError as e:  # noqa: F841
            pass
        pipcheck.wait()

        # installedpkgs = []
        #
        # for a preset set of named packages
        '''
        #pkgs = ['cloudmesh-common', 'cloudmesh-cmd5', 'cloudmesh.comet']
        for package in pkgs:
            # check version from pip
            pipcheck = subprocess.Popen(('pip', 'freeze'), stdout=subprocess.PIPE)
            try:
                output = subprocess.check_output(('grep', package), stdin=pipcheck.stdout)
                version = output.split("==")[1].strip()
                versions[package] = {"name": package,
                                     "version": version
                                     }
            except subprocess.CalledProcessError as e:
                pass
            pipcheck.wait()
        '''

        # __version__ not maintained in package file so this won't work
        '''
            try:
                print ("trying package |%s|" % package)
                try_module = __import__(package)
                print ("added one package into the list...")
                installedpkgs.append(package)
            except ImportError as e:
                print ("error importing |%s|" % package)
                pass

        #print (installedpkgs)
        for package in installedpkgs:
            versions[package] = {package: {"name": package,
                                           "version": str(package.__version__)
                                           }
                                 }
        print (versions)
        '''

        if arguments["--number"]:

            data = []
            for key, entry in versions.items():
                try:
                    d = {
                        'name': entry['package'],
                        'version': entry['VERSION']
                    }
                    data.append(d)
                except:
                    pass

            print(Printer.write(data,
                                output=arguments["--format"],
                                order=["name", "version"],
                                header=["Package", "Version"],
                                sort_keys="name"))
            return ""
        else:
            print(Printer.write(versions,
                                output=arguments["--format"],
                                order=["name", "package", "VERSION", "version",
                                       "source"],
                                sort_keys="name"))
        if arguments["--check"] in ["True"]:
            Shell.check_python()


# noinspection PyBroadException,PyUnusedLocal
def main():
    """cms.

    Usage:
      cms --help
      cms [--echo] [--debug] [--nosplash] [-i] [COMMAND ...]

    Arguments:
      COMMAND                  A command to be executed

    Options:
      --file=SCRIPT  -f  SCRIPT  Executes the script
      -i                 After start keep the shell interactive,
                         otherwise quit [default: False]
      --nosplash    do not show the banner [default: False]
    """

    def manual():
        print(main.__doc__)

    args = sys.argv[1:]

    arguments = {
        '--echo': '--echo' in args,
        '--help': '--help' in args,
        '--debug': '--debug' in args,
        '--nosplash': '--nosplash' in args,
        '-i': '-i' in args}

    echo = arguments["--echo"]
    if arguments['--help']:
        manual()
        sys.exit()

    for a in args:
        if a in arguments:
            args.remove(a)

    arguments['COMMAND'] = [' '.join(args)]

    commands = arguments["COMMAND"]
    # commands = list(arguments["COMMAND"])

    if len(commands) > 0:
        if ".cm" in commands[0]:
            arguments["SCRIPT"] = commands[0]
            commands = commands[1:]
        else:
            arguments["SCRIPT"] = None

        arguments["COMMAND"] = ' '.join(commands)
        if arguments["COMMAND"] == '':
            arguments["COMMAND"] = None

    # noinspection PySimplifyBooleanCheck
    if arguments['COMMAND'] == []:
        arguments['COMMAND'] = None

    # splash = not arguments['--nosplash']
    # debug = arguments['--debug']
    interactive = arguments['-i']
    script = arguments["SCRIPT"]
    command = arguments["COMMAND"]

    #
    # trick for the removal of the ", ' in the set command
    #
    if sys.platform != 'win32':
        if len(sys.argv) == 3:
            if sys.argv[1] == 'set' and "=" in sys.argv[2]:
                command = command.replace("\"", "")
                command = command.replace("=", "='", 1)
                command = command + "'"
        if len(sys.argv) >= 4:
            if sys.argv[1] == 'config' and sys.argv[2] == 'set' and "=" in sys.argv[3]:
                command = command.replace("\"", "")
                command = command.replace("=", "='", 1)
                command = command + "'"

    # import ctypes
    # ctypes.windll.kernel32.GetCommandLineA.restype = ctypes.c_char_p
    # print (ctypes.windll.kernel32.GetCommandLineA())

    # context = CloudmeshContext(
    #    interactive=interactive,
    #    debug=debug,
    #    echo=echo,
    #    splash=splash)

    cmd = CMShell()

    #    if script is not None:
    #        cmd.do_exec(script)

    try:
        if echo:
            print(cmd.prompt, command)
        if command is not None:
            cmd.precmd(command)
            stop = cmd.onecmd(command)
            cmd.postcmd(stop, command)
    except Exception as e:
        print("ERROR: executing command '{0}'".format(command))
        print(70 * "=")
        print(e)
        d = Default()
        trace = d["global", "trace"] == "True"
        trace = True
        Error.traceback(error=e, debug=True, trace=trace)
        d.close()
        print(70 * "=")

    if interactive or (command is None and script is None):
        cmd.cmdloop()


if __name__ == '__main__':
    main()



from cloudmesh.common.variables import Variables
from cloudmesh.shell.command import PluginCommand
from cloudmesh.shell.command import command


class DebugCommand(PluginCommand):
    # noinspection PyUnusedLocal

    @command
    def do_dryrun(self, args, arguments):
        """
        ::

            Usage:
                dryrun on
                dryrun off

            Description:

                dryrun on

                    sets the variable

                    dryrun=True

                dryrun off

                    sets the variable

                    dryrun=False

        """

        variables = Variables()

        if arguments.on:
            variables["dryrun"] = True

        elif arguments.off:
            variables["dryrun"] = False



from datetime import datetime

from cloudmesh.common.Shell import Shell
from cloudmesh.common.console import Console
from cloudmesh.common.variables import Variables
from cloudmesh.shell.command import PluginCommand
from cloudmesh.shell.command import command


class VarCommand(PluginCommand):
    # noinspection PyUnusedLocal

    @command
    def do_var(self, args, arguments):
        """
        ::

            Usage:
                var list
                var clear
                var delete NAME
                var NAME=VALUE
                var NAME
                var

            Arguments:
                NAME      the name of the variable
                VALUE     the value of the variable
                FILENAME  the filename of the variable

            Description:
                Manage persistent variables

                var NAME=VALUE
                   sets the variable with the name to the value
                   if the value is one of data, time, now it will be
                   replaced with the value at this time, the format will be
                    date    2017-04-14
                    time    11:30:33
                    now     2017-04-14 11:30:41
                It will wbe replaced accordingly

                The value can also refer to another variable name.
                In this case the current value will be copied in the named
                variable. As we use the $ sign it is important to distinguish
                shell variables from cms variables while using proper quoting.

                Examples include:

                   cms var a=\$b
                   cms var 'a=$b'
                   cms var a=val.b

                The previous command copy the value from b to a. The val command
                was added to avoid quoting.


        """  # noqa: W605

        if args == '':
            arguments["list"] = True

        variables = Variables()

        if arguments["NAME=VALUE"]:
            if '=' in arguments["NAME=VALUE"]:
                name, value = arguments["NAME=VALUE"].split("=", 1)

                if name in ['debug', 'trace', 'timer']:
                    value = str(str(value).lower() in ['1', 'on', 'true'])

            else:
                name = arguments["NAME=VALUE"]
                try:
                    value = variables[name]
                except:
                    value = None

        if arguments["clear"]:
            variables.clear()

        elif arguments["list"]:
            for name in variables:
                value = variables[name]
                print(name, "=", "'", value, "'", sep="")

        elif arguments.delete:
            del variables[arguments.NAME]

        elif name and not value:
            Console.error("variable {name} does not exist".format(name=name))

        elif name and not value:
            print(variables[arguments.NAME])

        elif name and value:
            if name in ["dryrun"]:
                if value.lower() in ["on", "true", "1", "t"]:
                    value = True
                elif value.lower() in ["off", "false", "0", "f"]:
                    value = True
                else:
                    Console.error(f"value of {name} must be True/False")
                    return ""
            elif value.startswith("cloudmesh."):
                try:
                    from cloudmesh.configuration.Config import Config
                    config = Config()
                except:
                    Console.error("cloudmesh configuration not loaded.")
                try:
                    value = config[value]
                except:
                    Console.error(f"problem reading {value}")
                    return ""
            elif value.startswith("!"):
                # cms set a=\!pwd
                command = value[1:]
                value = Shell.run(command)
            elif value.startswith("py "):
                # cms set a=\!pwd
                command = value.split("py ")[1]
                value = eval(command)

            elif value == "time":
                value = datetime.now().strftime("%H:%M:%S")
            elif value == "date":
                value = datetime.now().strftime("%Y-%m-%d")
            elif value == "now":
                value = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            elif value.startswith("value."):
                var = value.replace("value.", "")
                value = variables[var]
            elif value.startswith("$"):
                var = value.replace("$", "")
                value = variables[var]

            print(name, "=", "'", value, "'", sep="")
            variables[name] = value



import os
import sys
import time
from builtins import input

from cloudmesh.common.Shell import Shell
from cloudmesh.common.StopWatch import StopWatch
from cloudmesh.common.console import Console
from cloudmesh.shell.command import PluginCommand


class TerminalCommand(PluginCommand):
    from cloudmesh.shell.command import command

    # noinspection PyUnusedLocal,PyIncorrectDocstring
    @command
    def do_term(self, args, arguments):
        """
        ::

          Usage:
            term COMMAND

          Arguments:
            COMMAND   The command to execute in the terminal

           Opens a new terminal and executes the command in it
           The terminal stays openafter executing it in an interctive mode
           The terminal is started in the background

        """
        Shell.terminal(arguments.COMMAND)

        return ""

    # noinspection PyUnusedLocal,PyIncorrectDocstring
    @command
    def do_banner(self, args, arguments):
        """
        ::

          Usage:
            banner [-c CHAR] [-n WIDTH] [-i INDENT] [-r COLOR] TEXT...

          Arguments:
            TEXT...   The text message from which to create the banner
            CHAR      The character for the frame.
            WIDTH     Width of the banner
            INDENT    indentation of the banner
            COLOR     the color

            Options:
                -c CHAR   The character for the frame. [default: #]
                -n WIDTH  The width of the banner. [default: 70]
                -i INDENT  The width of the banner. [default: 0]
                -r COLOR  The color of the banner. [default: NORMAL]

            Prints a banner form a one line text message.
        """
        Console.ok("banner")
        n = int(arguments['-n'])
        c = arguments['-c']
        i = int(arguments['-i'])
        color = arguments['-r'].upper()

        line = ' '.join(arguments['TEXT'])

        Console.init()
        hline = i * " " + str((n - i) * c)
        Console.cprint(color=color, prefix="", message=hline)
        Console.cprint(color=color, prefix="", message=i * " " + c + " " + line)
        Console.cprint(color=color, prefix="", message=hline)

        return ""

    # noinspection PyUnusedLocal
    @command
    def do_stopwatch(self, args, arguments):
        """
        ::

          Usage:
            stopwatch start TIMER
            stopwatch stop TIMER
            stopwatch print [TIMER]
            stopwatch benchmark

          Arguments:
            TIMER  the name of the timer

          Description:
            THIS IS NOT YET WORKING
            starts and stops named timers and prints them
        """
        t = arguments.TIMER
        if arguments.start:
            StopWatch.start(t)
        elif arguments.stop:
            StopWatch.stop(t)
        elif arguments.benchmark:
            StopWatch.benchmark()
        elif arguments.print:
            t = arguments.TIMER
            if t is None:
                print(StopWatch.__str__)
            else:
                StopWatch.print("Timer " + t + ":", t)

    # noinspection PyUnusedLocal
    @command
    def do_clear(self, args, arguments):
        """
        ::

          Usage:
            clear



          Clears the screen."""

        sys.stdout.write(os.popen('clear').read())

    # noinspection PyUnusedLocal
    @command
    def do_sleep(self, args, arguments):
        """
        ::

          Usage:
            sleep SECONDS

          Clears the screen."""

        seconds = arguments["SECONDS"]
        time.sleep(float(seconds))

    # noinspection PyUnusedLocal
    @command
    def do_echo(self, args, arguments):
        """
        ::

          Usage:
            echo  [-r COLOR] TEXT

            Arguments:
                TEXT   The text message to print
                COLOR  the color

            Options:
                -r COLOR  The color of the text. [default: NORMAL]

            Prints a text in the given color
        """
        color = arguments["-r"] or "normal"
        color = color.upper()
        text = arguments["TEXT"]
        if color == "NORMAL":
            Console.msg(text)
        else:
            Console.cprint(color=color, prefix="", message=text)

        return ""

    @command
    def do_pause(self, arg, arguments):
        """
        ::

          Usage:
            pause [MESSAGE]

          Arguments:
            MESSAGE  message to be displayed

          Description:
            Displays the specified text then waits for the user to press RETURN.

        """

        if arguments["MESSAGE"] is None:
            arg = 'Press ENTER to continue'
        input(arg + '\n')

        return ""

    #
    # Echo
    #
    def set_verbose(self, on):
        # self.echo = on
        pass

    # noinspection PyAttributeOutsideInit
    def set_banner(self, banner):
        self.banner = banner

    # @command
    # def do_loglevel(self, args, arguments):
    #     """
    #     ::
    #
    #       Usage:
    #           loglevel
    #           loglevel critical
    #           loglevel error
    #           loglevel warning
    #           loglevel info
    #           loglevel debug
    #
    #           Shows current log level or changes it.
    #
    #           loglevel - shows current log level
    #           critical - shows log message in critical level
    #           error    - shows log message in error level including critical
    #           warning  - shows log message in warning level including error
    #           info     - shows log message in info level including warning
    #           debug    - shows log message in debug level including info
    #
    #       NOTE:
    #         NOT YET IMPLEMENTED
    #     """
    #
    #
    #     if arguments['debug']:
    #         self.loglevel = "DEBUG"
    #     elif arguments['error']:
    #         self.loglevel = "ERROR"
    #     elif arguments['warning']:
    #         self.loglevel = "WARNING"
    #     elif arguments['info']:
    #         self.loglevel = "INFO"
    #     elif arguments['critical']:
    #         self.loglevel = "CRITICAL"
    #     else:
    #         Console.ok("Log level: {0}".format(self.loglevel))
    #         return ""
    #     Console.ok ("Log level: {0} is set".format(self.loglevel))
    #
    #     config = Config()
    #     config["cloudmesh.logging.level"] = self.loglevel
    #     config.write("aaa.yaml")
    #     #config.write(filename=filename,
    #                   output="yaml",
    #                   attribute_indent="    ")
    #     return ""

    '''
    @command
    def do_verbose(self, args, arguments):
        """
        Usage:
            verbose (True | False)
            verbose

        NOTE: NOT YET IMPLEMENTED.
        If it sets to True, a command will be printed before execution.
        In the interactive mode, you may want to set it to False.
        When you use scripts, we recommend to set it to True.

        The default is set to False

        If verbose is specified without parameter the flag is
        toggled.

        """
        # if args == '':
        #    self.echo = not self.echo
        # else:
        #    self.echo = arguments['True']

        Console.error("verbose NOT YET IMPLEMENTED")

        return ""
        '''



import textwrap

from cloudmesh.common.Shell import Shell
from cloudmesh.common.console import Console
from cloudmesh.common.util import readfile
from cloudmesh.common.util import writefile
from cloudmesh.shell.command import PluginCommand
from cloudmesh.shell.command import command


class ManCommand(PluginCommand):
    # noinspection PyUnusedLocal

    def _convert_file(self, file, command, tag):
        tag_string = f"<!--{tag}-->"
        try:
            man = Shell.run(f"cms help {command}")

            # remove timer
            man = man.split("\nTimer: ")[0]

            readme = readfile("README.md")

            parts = readme.split(tag_string)

            content = []

            content.append(parts[0].strip())
            content.append("")
            content.append(tag_string)
            content.append("```")
            content.append(textwrap.dedent("\n".join(man.splitlines()[7:])))
            content.append("```")
            content.append(tag_string)
            try:
                content.append(parts[2])
            except:
                pass
        except Exception as e:
            content.append("")
            content.append(tag_string)
            content.append("```")
            content.append(e)
            content.append("```")
            content.append(tag_string)

        manpage = "\n".join(content)

        return manpage, man

    def _get_help(self, what):
        """
        prints the rst page of the command what
        :param what: the command
        :type what: string
        """
        h = None
        data = {"name": what}
        h = eval("self.do_{what}.__doc__".format(what=what))
        # noinspection PyUnboundLocalVariable
        data["help"] = h
        return data

    def _man_rst(self, data):
        result = [
            data['name'],
            "=" * len(data['name']),
            "",
            textwrap.dedent(data['help']).replace("::\n\n", ".. parsed-literal::\n\n").strip(),
            ""]
        return ("\n".join(result))

    def _man_txt(self, data):
        result = [
            "=" * 79,
            data['name'],
            "=" * 79,
            "",
            textwrap.dedent(data['help'].replace("::\n\n", "")).strip(),
            ""
        ]
        return ("\n".join(result))

    def _man_md(self, data):
        result = [
            "# " + data['name'],
            "",
            "```",
            textwrap.dedent(data['help'].replace("::\n\n", "")).strip(),
            "```",
            ""
        ]
        return ("\n".join(result))

    def _man_content(self, data, kind):
        result = ""
        if kind == "md":
            result = self._man_md(data)
        elif kind == "txt":
            result = self._man_txt(data)
        elif kind == "rst":
            result = self._man_rst(data)
        else:
            tmp = [
                data["name"],
                "=" * len(data["name"]),
                data["help"].replace("::\n\n", "", 1)
            ]
            result = "\n".join(tmp)

        return result

    def _print(self, name, data, kind, directory=None):
        content = self._man_content(data, kind)
        if directory is None:
            print(content)
        else:
            writefile(f"{directory}/{name}.{kind}", content)

    # noinspection PyUnusedLocal
    @command
    def do_man(self, arg, arguments):
        """
        ::

            Usage:
                man readme [-p] --toc [--file=FILE]
                man readme [-p] [--tag=TAG] [--file=FILE] COMMAND
                man [--dir=DIR] [--format=FORMAT] [--noheader]
                man COMMANDS... [--dir=DIR] [--format=FORMAT]

            Options:
                --toc        adds a table of content between the TOC tag
                -p           replacement in the file instead of stdout
                --noheader   no rst header
                --tag=TAG    the tag used to embed the manual
                             page [default: MANUAL]
                --file=FILE  the file for the replacement between the
                             tags [default: README.md]

            Arguments:
                COMMANDS   the command manual pages to be printed

            Description:
                man
                    Prints out the help pages
                man COMMAND
                    Prints out the help page for a specific command
        """

        cmds_doc = []
        cmds_undoc = []
        help_page = {}

        def get_manual_pages():

            names = self.get_names()

            for name in names:
                if name[:5] == 'help_':
                    help_page[name[5:]] = 1
            names.sort()
            # There can be duplicates if routines overridden
            prevname = ''
            for name in names:
                if name[:3] == 'do_':
                    if name == prevname:
                        continue
                    prevname = name
                    cmd = name[3:]
                    if cmd in help_page:
                        cmds_doc.append(cmd)
                        del help_page[cmd]
                    elif getattr(self, name).__doc__:
                        cmds_doc.append(cmd)
                    else:
                        cmds_undoc.append(cmd)

        arguments.kind = arguments["--format"] or "md"
        arguments.directory = arguments["--dir"]
        arguments.file = arguments["--file"] or "README.md"
        arguments.tag = arguments["--tag"] or "MANUAL"

        get_manual_pages()

        if arguments["--dir"]:
            d = arguments["--dir"]
            Shell.mkdir(d)

        if arguments["readme"] and arguments["--toc"]:

            in_place = arguments["-p"]
            if in_place:
                man = Shell.run(f"md_toc -p github {arguments.file}")
                print(man)
            else:
                man = Shell.run(f"md_toc github {arguments.file}")
                print(man)

        elif arguments["readme"]:

            in_place = arguments["-p"]

            manpage, old = self._convert_file(arguments.file,
                                              arguments.COMMAND,
                                              arguments.tag)
            if in_place and manpage != old:
                writefile(arguments.file, manpage)
            else:
                print(manpage)

        elif len(arguments.COMMANDS) == 0:

            for entry in cmds_doc:
                data = self._get_help(entry)
                self._print(entry, data, arguments.kind, arguments.directory)

        else:

            commands = arguments.COMMANDS

            for entry in commands:
                if entry in cmds_doc:

                    data = self._get_help(entry)
                    self._print(entry, data, arguments.kind, arguments.directory)

                else:
                    Console.error(f"Cloud not find man page for {entry}")

        return ""



from cloudmesh.shell.command import PluginCommand
from cloudmesh.shell.command import command
from cloudmesh.var.command.var import VarCommand


class SetCommand(PluginCommand):
    # noinspection PyUnusedLocal
    @command
    def do_set(self, args, arguments):
        """
        ::

          Usage:
            set list
            set clear
            set delete NAME
            set NAME=VALUE
            set NAME
            set

          Arguments:
            NAME      the name of the set variable
            VALUE     the value of the set variable
            FILENAME  the filename of the set variable

          Description:
            Manage persistent set variables

            set NAME=VALUE
               sets the set variable with the name to the value
               if the value is one of data, time, now it will be
               replaced with the value at this time, the format will be
                date    2017-04-14
                time    11:30:33
                now     2017-04-14 11:30:41
            It will wbe replaced accordingly

            The value can also refer to another setiable name.
            In this case the current value will be copied in the named
            setiable. As we use the $ sign it is important to distinguish
            shell set variables from cms setiables while using proper quoting.

            Examples include:

               cms set a=\$b
               cms set 'a=$b'
               cms set a=val.b

            The previous command copy the value from b to a. The val command
            was added to avoid quoting.

            A special variable is the variable verbose. It results in additional
            debug output. The following values apply

               0 - no additional output
              10 - most visible.
                   Prints all arguments from commands

        """  # noqa: W605

        c = VarCommand()
        c.do_var(args)



from cloudmesh.common.Printer import Printer
from cloudmesh.common.console import Console
from cloudmesh.common.default import Default

from cloudmesh.shell.command import PluginCommand
from cloudmesh.shell.command import command


class DefaultCommand(PluginCommand):
    # noinspection PyUnusedLocal
    @command
    def do_default(self, args, arguments):
        """
        ::

          Usage:
              default list [--context=CONTEXT] [--format=FORMAT]
              default delete --context=CONTEXT
              default delete KEY [--context=CONTEXT]
              default KEY [--context=CONTEXT]
              default KEY=VALUE [--CONTEXT=CONTEXT]

          Arguments:
            KEY    the name of the default
            VALUE  the value to set the key to

          Options:
             --context=CONTEXT    the name of the context
             --format=FORMAT  the output format. Values include
                              table, json, csv, yaml.

          Description:
            Cloudmesh has the ability to manage easily multiple
            clouds. One of the key concepts to manage multiple clouds
            is to use defaults for the cloud, the images, flavors,
            and other values. The default command is used to manage
            such default values. These defaults are used in other commands
            if they are not overwritten by a command parameter.

            The current default values can by listed with

                default list --all

            Via the default command you can list, set, get and delete
            default values. You can list the defaults with

               default list

            A default can be set with

                default KEY=VALUE

            To look up a default value you can say

                default KEY

            A default can be deleted with

                default delete KEY

            To be specific to a cloud you can specify the name of the
            cloud with the --cloud=CLOUD option. The list command can
            print the information in various formats iv specified.

          Examples:
            default list --all
                lists all default values

            default list --cloud=kilo
                lists the defaults for the cloud with the name kilo

            default image=xyz
                sets the default image for the default cloud to xyz

            default image=abc --cloud=kilo
                sets the default image for the cloud kilo to xyz

            default image
                list the default image of the default cloud

            default image --cloud=kilo
                list the default image of the cloud kilo

            default delete image
                deletes the value for the default image in the
                default cloud

            default delete image --cloud=kilo
                deletes the value for the default image in the
                cloud kilo

        """
        # print(arguments)
        # print (">", args, "<")

        context = arguments["--context"] or "global"

        defaults = Default()

        if arguments["delete"]:

            key = arguments.KEY
            if key is None:
                del defaults[context]
            else:
                del defaults[context, key]
            # if not result :
            #    Console.error("default {} not present".format(key))
            # else:
            #    Console.ok("Deleted key {} for cloud {}. ok.".format(key,
            #                                                      cloud))
            return ""

        if arguments.KEY is not None:
            if "=" not in arguments.KEY:
                print(arguments.KEY, "=", defaults[context, arguments.KEY],
                      sep='')
            else:
                key, value = arguments["KEY"].split("=", 1)
                defaults[context, key] = value
            return ""

        if arguments["list"]:

            output_format = arguments["--format"] or defaults[
                "global", "output"] or 'table'

            if defaults is None:
                Console.error("No default values found")
            else:

                data = []
                d = defaults.__dict__()
                for context in d:
                    for key in d[context]:
                        data.append({"context": context,
                                     "key": key,
                                     "value": d[context][key]})
                print(Printer.list(data, output=output_format))
            return ""



from cloudmesh.shell.command import PluginCommand
from cloudmesh.shell.command import command


class DebugCommand(PluginCommand):
    # noinspection PyUnusedLocal

    @command
    def do_debug(self, args, arguments):
        """
        ::

            Usage:
                debug on
                debug off


            Description:

                debug on

                    sets the variables

                    debug=True
                    trace=True
                    verbose=10
                    timer=True

                debug off

                    sets the variables

                    debug=False
                    trace=False
                    verbose=0
                    timer=False


        """
        print(args, arguments)
        self.set_debug(args)



import shlex
import textwrap

from cloudmesh.common.console import Console
from cloudmesh.common.dotdict import dotdict
from docopt import docopt


class PluginCommand(object):
    pass


class CloudPluginCommand(object):
    pass


class ShellPluginCommand(object):
    pass


class HPCPluginCommand(object):
    pass


class CometPluginCommand(object):
    pass


def map_parameters(arguments, *args):
    """
    This command is useful to map parameters with -- to regular argument dicts
    for easier processing.

    :param arguments:
    :param args:
    :return:

    an example is

    map_parameters(arguments,
                       'active',
                       'cloud')

    where --active=ACTIVE is mapped to arguments["active"]
    and  --cloud=CLOUD is mapped to arguments["cloud"]

    as arguments is a dotdict, they can than for example be called as

    arguments.cloud

    """
    for arg in args:
        if arg in ['clear', 'copy', 'fromkeys', 'get', 'items', 'keys', 'pop',
                   'popitem', 'setdefault', 'update', 'values', 'format',
                   'type']:
            Console.error(f'`{arg}` is predefined method.'
                          f' Use `arguments["--{arg}"]` in your code')
            raise ValueError(f"{arg} already used in arguments")
        elif arg in arguments:
            Console.error(f'`{arg}` is already used in arguments.'
                          f' Use `arguments["--{arg}"]` in your code')
            raise ValueError(f"{arg} already used in arguments")
        else:
            flag = "--" + arg
            if flag in arguments:
                value = arguments[flag]
            else:
                value = None
            arguments[arg] = value


# noinspection PySingleQuotedDocstring,PyUnusedLocal
def command(func):
    '''
    A decorator to create a function with docopt arguments.
    It also generates a help function

    @command
    def do_myfunc(self, args):
        """ docopts text """
        pass

    will create

    def do_myfunc(self, args, arguments):
        """ docopts text """
        ...

    def help_myfunc(self, args, arguments):
        ... prints the docopt text ...

    :param func: the function for the decorator
    '''
    # classname = inspect.getouterframes(inspect.currentframe())[1][3]
    name = func.__name__
    # help_name = name.replace("do_", "help_")
    doc = textwrap.dedent(func.__doc__)

    def new(instance, args):
        # instance.new.__doc__ = doc
        # noinspection PyUnusedLocal
        try:
            argv = shlex.split(args)
            # pprint(argv)
            arguments = dotdict(docopt(doc, help=True, argv=argv))
            # pprint(arguments)
            # verbose = int(Variables()["verbose"] or 0)
            # if verbose > 9:
            #    s = pformat(arguments)
            #    banner(s, label="Arguments", color="BLUE")
            func(instance, args, arguments)
        except SystemExit as e:  # noqa: F841
            if args not in ('-h', '--help'):
                if "::" in doc:
                    usage = textwrap.dedent(doc.split("::")[1])
                else:
                    usage = doc

                usage = textwrap.dedent(doc.split("Usage:")[1])
                print()
                print("Usage:")

                for line in usage.splitlines():
                    if ":" in line:
                        kind = line.split(":")[0]
                        if kind in ["Arguments", "Options", "Example",
                                    "Descriptiom"]:
                            break
                    print(line)
                Console.error(
                    "Could not execute the command. Please check usage with")
                print()
                Console.msg("    cms help", name.replace("do_", ""))
                print()
                # print (args)
                # print(e)
                # print(doc)

    new.__doc__ = doc
    return new


# noinspection PySingleQuotedDocstring,PyUnusedLocal
def basecommand(func):
    '''
    A decorator to create a function with docopt arguments.
    It also generates a help function

    @command
    def do_myfunc(self, args):
        """ docopts text """
        pass

    will create

    def do_myfunc(self, args, arguments):
        """ docopts text """
        ...

    def help_myfunc(self, args, arguments):
        ... prints the docopt text ...

    :param func: the function for the decorator
    '''
    # classname = inspect.getouterframes(inspect.currentframe())[1][3]
    # name = func.__name__
    # help_name = name.replace("do_", "help_")
    doc = textwrap.dedent(func.__doc__)

    def new(instance, args):
        # instance.new.__doc__ = doc
        # noinspection PyUnusedLocal
        try:
            # print("ARGS", args)
            argv = shlex.split(args)
            # print ("ARGV", argv)
            arguments = docopt(doc, help=True, argv=argv)
            func(instance, args, arguments)
        # except docopt.DocoptExit as e:
        except Exception as e:  # noqa: F841
            import traceback
            import sys
            traceback.print_exc(file=sys.stdout)
            if args not in ('-h', '--help'):
                print(args)
                Console.error("Could not execute the command.")
                Console.error("Check usage..")
            print(doc)

        except SystemExit as e:
            if args not in ('-h', '--help'):
                print(args)
                Console.error("Could not execute the command.")
                Console.error("Check usage..")
                print(e)
            print(doc)

    new.__doc__ = doc
    return new
